#include "super_frame.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include "utils/log.h"
#include "utils/bit_reader.h"
#include "utils/crc.h"
#include "aac.h"
#include "coding/reed_solomon.h"
#include "coding/firecode.h"

super_frame *new_super_frame(int subchannel_size,
    void (*send_audio)(void *, void *, int),
    void (*send_audio_config)(void *, int, int),
    void *send_audio_data)
{
  super_frame *ret = calloc(1, sizeof(super_frame));
  if (ret == NULL) abort();

  ret->subchannel_index = subchannel_size / 8;
  ret->audio_super_frame_size = ret->subchannel_index * 110;

  ret->cif_start = -1;

  ret->data = malloc(ret->audio_super_frame_size + ret->subchannel_index * 10);
  if (ret->data == NULL) abort();

  ret->frame_size = (ret->audio_super_frame_size + ret->subchannel_index * 10) / 5;

  ret->aac_decoder = aac_new_decoder();

  ret->send_audio = send_audio;
  ret->send_audio_config = send_audio_config;
  ret->send_audio_data = send_audio_data;

  reed_solomon_init(&ret->rs, (1<<8) + (1<<4) + (1<<3) + (1<<2) + 1, 5, 0, 110);

  return ret;
}

void free_super_frame(super_frame *s)
{
  free_aac_decoder(s->aac_decoder);
  free(s->data);
  reed_solomon_free(&s->rs);
  free(s);
}

/* 102.563 6.2, 6.3, 6.4, 6.5
 * (s = subchannel_index)
 * C matrix C[s][110] = (say s=11)
 *     A0  A11 A22 .. A1199
 *     A1  A12 A23 .. A1200
 *            ...
 *     A10 A21 A32 .. A1209
 * data[i] = Ai
 * P[s][10] transmitted in order after A as:
 *    P[0][0]  P[1][0] .. P[9][0] P[10][0]
 *    P[0][1]  ..
 *            ...
 *    P[0][9]  ..
 * so to read P[0], read data[pos], data[pos+11], data[pos+11*2], ...
 */

int do_rs = 1;

void reed_solomon(reed_solomon_t *rs, unsigned char *data, int subchannel_index)
{
  if (!do_rs) return;

  unsigned char c[110 + 10];

  for (int i = 0; i < subchannel_index; i++) {
    for (int j = 0; j < 110; j++)
      c[j] = data[i + j * subchannel_index];
    for (int j = 0; j < 10; j++)
      c[110 + j] = data[110 * subchannel_index + i + j * subchannel_index];

    int error_count = reed_solomon_decode(rs, c);
    debug("reed-solomon corrected %d bytes\n", error_count);

    for (int count = 0; count < error_count; count++) {
      int j = rs->correction[count].pos;
      if (j < 110)
        data[i + j * subchannel_index] = rs->correction[count].corrected_value;
      else
        data[110 * subchannel_index + i + (j-110) * subchannel_index] = rs->correction[count].corrected_value;
    }
  }
}

/* code based on ETSI TS 102 563 V2.1.1 */
static void audio_decode_super_frame(super_frame *s)
{
  struct {
    int num_aus;
    int au_start;
    int sample_rate;
  } conf[4] = {
    { 4,  8, 32 },
    { 2,  5, 16 },
    { 6, 11, 48 },
    { 3,  6, 24 },
  };

  /* fire decode - resync if failure */
  int computed_crc = crc((char *)s->data+2, 9, 0x1782f, 16, 0, CRC_NO_COMPLEMENT);
  int received_crc = (s->data[0] << 8) | s->data[1];
  if (computed_crc != received_crc) {
    if (fire_decode(s->data)) {
      /* bad crc, unrecoverable? resync */
      s->cif_start = -1;
      return;
    }
  }

  reed_solomon(&s->rs, s->data, s->subchannel_index);

  bit_reader b;
  int dac_rate;
  int sbr_flag;
  int aac_channel_mode;
  int ps_flag;
  int mpeg_surround_config;
  int num_aus;
  int n;
  int au_start[6+1];
  int au_size[6];
  int sample_rate;

  init_bit_reader(&b, s->data, s->audio_super_frame_size);

  get_bits(&b, 16);                           /* header_firecode */
  get_bits(&b, 1);                            /* rfa */
  dac_rate             = get_bits(&b, 1);
  sbr_flag             = get_bits(&b, 1);
  aac_channel_mode     = get_bits(&b, 1);
  ps_flag              = get_bits(&b, 1);
  mpeg_surround_config = get_bits(&b, 3);

  debug("superframe config: dac_rate %d sbr_flag %d aac_channel_mode %d ps_flag %d mpeg_surround_config %d\n",
        dac_rate, sbr_flag, aac_channel_mode, ps_flag, mpeg_surround_config);

  n = (dac_rate << 1) | sbr_flag;

  num_aus     = conf[n].num_aus;
  au_start[0] = conf[n].au_start;
  sample_rate = conf[n].sample_rate;

  for (n = 1; n < num_aus; n++) {
    au_start[n] = get_bits(&b, 12);
    debug("au start[%d] = %d\n", n, au_start[n]);
    if (au_start[n] <= au_start[n-1] + 2
       || au_start[n] >= s->audio_super_frame_size - 2) {
      debug("bad superframe's header, skipping\n");
      s->bad_count++;
      if (s->bad_count == 5) {
        s->bad_count = 0;
        s->cif_start = -1;
      }
      return;
    }
  }

  au_start[num_aus] = s->audio_super_frame_size;

  for (n = 0; n < num_aus; n++)
    au_size[n] = au_start[n+1] - au_start[n] - 2;

  /* if all AUs have bad CRC then superframe is fully bad and then we increase
   * bad_count, but if one AU is good, then we clear bad_count
   */
  bool fully_bad = true;

  /* check CRCs */
  for (n = 0; n < num_aus; n++) {
    int received_crc = (s->data[au_start[n] + au_size[n]] << 8) |
                        s->data[au_start[n] + au_size[n] + 1];
    int computed_crc = crc((char *)&s->data[au_start[n]], au_size[n],
                           0x11021, 16, 0xffff, CRC_COMPLEMENT);
    if (received_crc != computed_crc) {
      debug("superframe: bad crc (received %2.2x computed %2.2x), skip au %d (size %d)\n",
             received_crc, computed_crc, n, au_size[n]);
      au_size[n] = 0;
    } else
      fully_bad = false;
  }

  if (fully_bad) {
    s->bad_count++;
    if (s->bad_count == 5) {
      s->bad_count = 0;
      s->cif_start = -1;
      return;
    }
  } else
    s->bad_count = 0;

  for (n = 0; n < num_aus; n++) {
    debug("au start %d size %d\n", au_start[n], au_size[n]);
  }

  aac_config_t aac_config = aac_set_config(s->aac_decoder, sample_rate, aac_channel_mode);
  if (aac_config.samplerate != -1)
    s->send_audio_config(s->send_audio_data, aac_config.samplerate, aac_config.channels);

  for (n = 0; n < num_aus; n++) {
    int outlen;
    short *out;
    if (au_size[n] == 0)
      continue;
    out = decode_au(s->aac_decoder, s->data + au_start[n], au_size[n], &outlen);
    s->send_audio(s->send_audio_data, out, outlen*2);
  }
}

void audio_receive_frame(super_frame *s, int cif_id, 
                         unsigned char *data, int len)
{
  if (len != s->frame_size) {
    error("%s:%d:%s: fatal, bad size\n", __FILE__, __LINE__, __FUNCTION__);
    return;
    exit(1);
  }

  /* check continuity of cif_id */
  if (cif_id != s->next_expected_cif) {
    /* we need to resync */
    s->cif_start = -1;
  }

  s->next_expected_cif = (cif_id + 1) % 5000;

  /* resync in progress? */
  if (s->cif_start == -1) {
    /* firecode: check crc, if different, try to fire_decode() */
    int computed_crc = crc((char *)data+2, 9, 0x1782f, 16, 0, CRC_NO_COMPLEMENT);
    int received_crc = (data[0] << 8) | data[1];
    if (computed_crc != received_crc) {
      if (fire_decode(data)) return;
    }

    /* sync done */
    s->cif_start = cif_id;
    s->pos = 0;
  }

  memcpy(s->data + s->pos * s->frame_size, data, len);
  s->pos++;

  /* check full super frame received */
  if (s->pos == 5) {
    audio_decode_super_frame(s);
    s->pos = 0;
  }
}
